/* CDCC v2 1.00 (c) Copyright 1996 William Glozer  */
/* ----------------------------------------------- */
/* Last revision 04.19.96 by Ananda                */

/* New CDCC for BitchX and any other clients that deserve to run it :) */
/* Note that I did use a lot of code/ideas from a copy of CDCC written */
/* for BitchX by panasync, so thanks to him for alot of the code and   */
/* ideas :)  I would appreciate any bugs reported to me as soon as is  */
/* possible, and cdcc.c + cdcc.h for any mods you do... if you modify  */
/* it for your client, I can add those mods as #ifdefs so the next ver */
/* will work with your client and have all the new stuff.  -Ananda '96 */
/* Modifed even more by panasync (edwac@sasknet.sk.ca) to interface    */
/* cleanly and nicely with BitchX. Blame all bugs on me instead of     */
/* Ananda */

#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "irc.h"
#include "ircaux.h"
#include "ignore.h"
#include "ctcp.h"
#include "hook.h"
#include "dcc.h"
#include "flood.h"
#include "screen.h"
#include "parse.h"
#include "output.h"
#include "input.h"
#include "vars.h"
#include "whois.h"

#include "cdcc.h"
#define BITCHX
#define CDCC_FLUD

/* external ircII stuff */
extern int dcc_active_count;
extern void dcc_filesend(char *);
extern void dcc_resend(char *);
int l_timer _((char *, char *));

/* local commands */
local_cmd local[] = {
	{ "HELP",	l_help,		"get help on CDCC commands" },
	{ "OFFER",	l_offer,	"add a pack to the offer list" },
	{ "DOFFER",	l_doffer,	"remove pack from the offer list" },
	{ "LIST",	l_list,		"list the packs you have offered" },
	{ "PLIST",	l_plist,	"publicly list your offered packs" },
	{ "NOTICE",	l_notice,	"notify the channel of offered packs" },
	{ "QUEUE",	l_queue,	"view entries in the send queue" },
	{ "SAVE",	l_save,		"save your offerlist to cdcc.save" },
	{ "SEND",	l_send,		"send a pack to user" },
	{ "RESEND",	l_resend,	"re-send a pack to user" },
	{ "LOAD",	l_load,		"load packs saved to cdcc.save" },
	{ "TIMER",	l_timer,	"public list timer" },
	{ "",		NULL,		"" } 
};
#define NUM_LOCAL (sizeof(local) / sizeof(local_cmd))

/* remote commands */
remote_cmd remote[] = {
	{ "HELP",	r_help,		"get help on CDCC commands" },
	{ "RESEND",	r_rsend,	"have CDCC re-send you pack #N"},
	{ "SEND",	r_send,		"have CDCC send you pack #N" },
	{ "LIST",	r_list,		"get a list of offered packs" },
	{ "",		NULL,		"" }, 
};
#define NUM_REMOTE (sizeof(remote) / sizeof(remote_cmd))

/* ahh yes, global variables :( well, interfacing with ircII is a pain in  */
/* the ass, and I couldn't figure out a better way... more than one of my  */
/* routines need these... the static will keep them from being used by any */
/* functions outside of cdcc.c                                             */

static int numpacks = 0;
static pack *offerlist = NULL;
static pack *newpack;

static queue *queuelist = NULL;
static int numqueue = 0;
static int ptimer = 0;

/* parse a users CDCC command */
void cdcc(char *command, char *args, char *blah)
{
	int i;
	char *cmd, *rest;
	
	cmd = next_arg(args, &args);
	rest = next_arg(args, &args);
	if (!cmd) {
		put_it("*** \002CDCC\002: try /cdcc help");
		return;
	}
	
	for (i = 0; *local[i].name; i++) {
		if (!my_stricmp(local[i].name, cmd)) {  
			local[i].function(rest, args);
			return;
		}
	}
	put_it("*** \002CDCC\002: unknown command \002%s\002", cmd);
	return;
}
			
/* parse a remote message CDCC command */
char *msgcdcc(char *from, char *to, char *args)
{
	int i;
	char *cdcc, *rest, *cmd, *temp = NULL, *q;

	malloc_strcpy(&temp, args);
	
	q = temp;
	cdcc = next_arg(temp, &temp);
	if (!cdcc || my_strnicmp(++cdcc, "DCC", 3)) {
	 	new_free(&q);
	 	return args;
	 } 
	
	if ((check_ignore(from, FromUserHost, IGNORE_CDCC) == IGNORED))
		return NULL;
		
	if (!check_flooding(from, CDCC_FLOOD, args))
		return NULL;
	cmd = next_arg(temp, &temp);
	if (!cmd) {
		new_free(&q);
		return args;
	}

	rest = temp;

	for (i = 0; *remote[i].name; i++) {
		if (!my_stricmp(cmd, remote[i].name)) {
			if (get_int_var(CLOAK_VAR))
				return args;
			remote[i].function(from, rest);
			new_free(&q);
			return NULL;
		}
	}
	send_to_server("NOTICE %s :try /ctcp %s cdcc help",
		from, nickname);
	new_free(&q);
	return args;
}
 
/* parse a remote ctcp CDCC command */
char *do_bdcc(CtcpEntry *ctcp, char *from, char *to, char *args)
{
	int i;
	char *rest, *cmd, *temp = NULL, *q;
	
	if ((check_ignore(from, FromUserHost, IGNORE_CDCC) == IGNORED))
		return NULL;
		
	if (!check_flooding(from, CDCC_FLOOD, args))
		return NULL;
	malloc_strcpy(&temp, args);
	q = temp;
 	
	cmd = next_arg(temp, &temp);
	if (!cmd) {
		new_free(&q);
		return NULL;
	}
	rest = temp;
	for (i = 0; *remote[i].name; i++) 
	{
		if (!my_stricmp(cmd, remote[i].name)) 
		{
			if (get_int_var(CLOAK_VAR))
				return NULL;
			remote[i].function(from, rest);
			new_free(&q);
			return NULL;
		}
	}
	send_to_server("NOTICE %s :try /ctcp %s cdcc help",
		from, nickname);
	new_free(&q);
	 
	return NULL;
}

int l_send(char *args, char *rest)
{
	pack *ptr;
	char *temp = NULL, *file = NULL, *dccinfo = NULL, *q;
	int packnum, maxdcc, maxqueue;
	if (!numpacks) {
		put_it("*** \002CDCC\002: you have no packs offered");
		return 0;
	}
	if (!args || !*args)
	{
		put_it("*** \002CDCC\002: Usage:  /cdcc send nickname #");
		return 0;
	}
	temp = strchr(rest, '#');
	if (!temp)
		temp = rest;
	else
		temp++;

	packnum = atoi(temp);
	if (!packnum || packnum > numpacks) {
		put_it("*** \002CDCC\002: invalid pack number");
		return 0;
	}
	maxdcc = get_int_var(DCC_SEND_LIMIT_VAR);
	maxqueue = get_int_var(DCC_QUEUE_LIMIT_VAR);
	
	for (ptr = offerlist; ptr; ptr = ptr->next) {
		if (ptr->num == packnum)
			break;
	}
	
	if ((maxdcc - dcc_active_count) < ptr->numfiles || dcc_active_count >= maxdcc) {
		if (numqueue >= maxqueue) {
			put_it("*** \002CDCC\002: all dcc and queue slots full");
			return 0;
		}
		add_to_queue(args, ptr);			
		return 0;
	}
	
	put_it("*** \002CDCC\002: sending %s%s%s pack #\002%d\002 (\002%d\002 file%s)",
		UND_TOG_STR, args, UND_TOG_STR, packnum, ptr->numfiles,
		ptr->numfiles == 1 ? "" : "s");
	malloc_strcpy(&file, ptr->file);
	q = file;
	for (temp = next_arg(file, &file); temp && *temp; temp = next_arg(file, &file)) {
		malloc_sprintf(&dccinfo, "%s %s", args, temp);
		dcc_filesend(dccinfo);
	}
	new_free(&dccinfo);
	new_free(&q);
	ptr->gets++;
	
	return 0;
}

int l_resend(char *args, char *rest)
{
	pack *ptr;
	char *temp = NULL, *file = NULL, *dccinfo = NULL, *q;
	int packnum, maxdcc, maxqueue;
	if (!numpacks) {
		put_it("*** \002CDCC\002: you have no packs offered");
		return 0;
	}
	if (!args || !*args)
	{
		put_it("*** \002CDCC\002: Usage:  /cdcc send nickname #");
		return 0;
	}
	temp = strchr(rest, '#');
	if (!temp)
		temp = rest;
	else
		temp++;

	packnum = atoi(temp);
	if (!packnum || packnum > numpacks) {
		put_it("*** \002CDCC\002: invalid pack number");
		return 0;
	}
	maxdcc = get_int_var(DCC_SEND_LIMIT_VAR);
	maxqueue = get_int_var(DCC_QUEUE_LIMIT_VAR);
	
	for (ptr = offerlist; ptr; ptr = ptr->next) {
		if (ptr->num == packnum)
			break;
	}
	
	if ((maxdcc - dcc_active_count) < ptr->numfiles || dcc_active_count >= maxdcc) {
		if (numqueue >= maxqueue) {
			put_it("*** \002CDCC\002: all dcc and queue slots full");
			return 0;
		}
		add_to_queue(args, ptr);			
		return 0;
	}
	
	put_it("*** \002CDCC\002: sending %s%s%s pack #\002%d\002 (\002%d\002 file%s)",
		UND_TOG_STR, args, UND_TOG_STR, packnum, ptr->numfiles,
		ptr->numfiles == 1 ? "" : "s");
	malloc_strcpy(&file, ptr->file);
	q = file;
	for (temp = next_arg(file, &file); temp && *temp; temp = next_arg(file, &file)) {
		malloc_sprintf(&dccinfo, "%s %s", args, temp);
		dcc_resend(dccinfo);
	}
	new_free(&dccinfo);
	new_free(&q);
	ptr->gets++;
	
	return 0;
}

/*resends a pack to the requestee*/
/*Added by Wicked Angel: wangel@wgrobez1.remote.louisville.edu*/
/*It wasn't hard ... but hey ... it seemed like a good idea :>*/
int r_rsend(char *from, char *args)
{
	pack *ptr;
	char *temp = NULL, *file = NULL, *dccinfo = NULL, *q;
	int packnum, maxdcc, maxqueue;
	
	temp = strchr(args, '#');
	if (!temp)
		temp = args;
	else
		temp++;

	packnum = atoi(temp);
	if (!packnum || packnum > numpacks) {
		send_to_server("NOTICE %s :\002CDCC\002: invalid pack number",
			from);
		return 0;
	}
	maxdcc = get_int_var(DCC_SEND_LIMIT_VAR);
	maxqueue = get_int_var(DCC_QUEUE_LIMIT_VAR);
	
	for (ptr = offerlist; ptr; ptr = ptr->next) {
		if (ptr->num == packnum)
			break;
	}
	
	if ((maxdcc - dcc_active_count) < ptr->numfiles || dcc_active_count >= maxdcc) {
		if (numqueue >= maxqueue) {
			send_to_server("NOTICE %s: \002CDCC\002: all dcc and queue slots full, sorry",
				from);
			return 0;
		}
		add_to_queue(from, ptr);			
		return 0;
	}
	
	put_it("*** \002CDCC\002: resending %s%s%s pack #\002%d\002 (\002%d\002 file%s)",
		UND_TOG_STR, from, UND_TOG_STR, packnum, ptr->numfiles,
		ptr->numfiles == 1 ? "" : "s");
	malloc_strcpy(&file, ptr->file);
	q = file;
	for (temp = next_arg(file, &file); temp && *temp; temp = next_arg(file, &file)) {
		malloc_sprintf(&dccinfo, "%s %s", from, temp);
		dcc_resend(dccinfo);
	}
	new_free(&dccinfo);
	new_free(&q);
	ptr->gets++;
	return 0;
}

/* send a pack to the remote user */
int r_send(char *from, char *args)
{
	pack *ptr;
	char *temp = NULL, *file = NULL, *dccinfo = NULL, *q;
	int packnum, maxdcc, maxqueue;
	
	temp = strchr(args, '#');
	if (!temp)
		temp = args;
	else
		temp++;

	packnum = atoi(temp);
	if (!packnum || packnum > numpacks) {
		send_to_server("NOTICE %s :\002CDCC\002: invalid pack number",
			from);
		return 0;
	}
	maxdcc = get_int_var(DCC_SEND_LIMIT_VAR);
	maxqueue = get_int_var(DCC_QUEUE_LIMIT_VAR);
	
	for (ptr = offerlist; ptr; ptr = ptr->next) {
		if (ptr->num == packnum)
			break;
	}
	
	if ((maxdcc - dcc_active_count) < ptr->numfiles || dcc_active_count >= maxdcc) {
		if (numqueue >= maxqueue) {
			send_to_server("NOTICE %s: \002CDCC\002: all dcc and queue slots full, sorry",
				from);
			return 0;
		}
		add_to_queue(from, ptr);			
		return 0;
	}
	
	put_it("*** \002CDCC\002: sending %s%s%s pack #\002%d\002 (\002%d\002 file%s)",
		UND_TOG_STR, from, UND_TOG_STR, packnum, ptr->numfiles,
		ptr->numfiles == 1 ? "" : "s");
	malloc_strcpy(&file, ptr->file);
	q = file;
	for (temp = next_arg(file, &file); temp && *temp; temp = next_arg(file, &file)) {
		malloc_sprintf(&dccinfo, "%s %s", from, temp);
		dcc_filesend(dccinfo);
	}
	new_free(&dccinfo);
	new_free(&q);
	ptr->gets++;
	
	return 0;
}
			

/* remote pack list */
int r_list(char *from, char *args)
{
	pack *ptr;
	char size[11];
	int once = 0;
		
	for (ptr = offerlist; ptr; ptr = ptr->next) {
#ifdef BITCHX
		if (!once && do_hook(CDCC_PREPACK_LIST, "%s %s %d %d %d %d %d", from, nickname, numpacks, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR), numqueue, get_int_var(DCC_QUEUE_LIMIT_VAR)))
		{
			send_to_server("NOTICE %s :Files Offered: /ctcp %s CDCC send #N for pack N", from, nickname);
			if (get_int_var(DCC_SEND_LIMIT_VAR))
				send_to_server("NOTICE %s :    [%d packs %d/%d slots open]", from, 
					numpacks, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR));
			else
				send_to_server("NOTICE %s :    [%d packs]", from, numpacks);
		}
#endif
	 	if (ptr->size / 1024 > 999)
			sprintf(size, "\002%3.2g\002mb",
				(double) (ptr->size / 1024) / 1024);
		else
			sprintf(size, "\002%3.2g\002kb", (double) ptr->size / 1024); 
				
#ifdef BITCHX
		if (do_hook(CDCC_PACK_LIST, "%s %d %d %lu %d %s", 
			from, ptr->num, ptr->numfiles, ptr->size, ptr->gets, ptr->desc))
		{
#endif
			send_to_server("NOTICE %s :#%d \037(\037%10s\037:\037\002%4d\002 get%s\037)\037 %-40.40s",
				from, ptr->num, size, ptr->gets, ptr->gets == 1 ? " " : "s", 
				ptr->desc ? ptr->desc : "no description");
#ifdef BITCHX
		}
		once++;
#endif
	}
#ifdef BITCHX
	if (once)
		(void) do_hook(CDCC_POSTPACK_LIST, "%s %s %d %d %d %d %d", from, nickname, numpacks, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR), numqueue, get_int_var(DCC_QUEUE_LIMIT_VAR));
#endif
	return 0;
}		

/* remote help display  */
int r_help(char *from, char *args)
{
	int i;
#ifdef CDCC_FLUD
	for (i = 0; *remote[i].name; i++) {
		send_to_server("NOTICE %s :\002CDCC\002: %-4s - %s",
			from, remote[i].name, remote[i].help);
	}
#endif
	return 0;
}


/* remove a pack or all packs from the offerlist */
int l_doffer(char *args, char *rest)
{
	if (!numpacks) {
		put_it("*** \002CDCC\002: you have no packs offered");
		return 0;
	}
	l_list(NULL, NULL);
	add_wait_prompt("remove pack [* for all packs]: ", del_pack, "", WAIT_PROMPT_LINE);
	return 0;
}

/* localy list the packs you have offered */
int l_list(char *args, char *rest)
{
	pack *ptr;
	char size[11];

	if (!numpacks) {
		put_it("*** \002CDCC\002: you have no packs offered");
		return 0;
	}
	
	for (ptr = offerlist; ptr; ptr = ptr->next) {
		if (ptr->size / 1024 > 999)
			sprintf(size, "\002%3.2f\002mb",
				(double) (ptr->size / 1024) / 1024);
		else
			sprintf(size, "\002%3.2f\002kb", (double) ptr->size / 1024); 
	
		put_it("#%-2d \002%3d\002 file%s %10s \002%4d\002 get%s %-39.39s",
			ptr->num, ptr->numfiles, ptr->numfiles == 1 ? " " : "s",
			size, ptr->gets, ptr->gets == 1 ? " " : "s", ptr->desc);
	}
	return 0;
}

/* add a pack to the offer list */
int l_offer(char *args, char *rest)
{
	add_wait_prompt("add file(s) to pack: ", add_files, "", WAIT_PROMPT_LINE);
	return 0;
}

/* display the offerlist to current channel */
int l_plist(char *args, char *rest)
{
	pack *ptr;
	char *chan = NULL, *string = NULL;
	char size[20];
	int maxdccs, blocksize, maxqueue;
	
	if (!get_channel_by_refnum(0) || !numpacks) {
		put_it("*** \002CDCC\002: you %s",
			numpacks ? "are not on a channel!" :
				   "have no packs offered!");
		return 0;
	}
	
	malloc_strcpy(&chan, get_channel_by_refnum(0));

	maxdccs = get_int_var(DCC_SEND_LIMIT_VAR);
	blocksize = get_int_var(DCC_BLOCK_SIZE_VAR);
	maxqueue = get_int_var(DCC_QUEUE_LIMIT_VAR);

#ifdef BITCHX
	if (do_hook(CDCC_PREPACK_LIST, "%s %s %d %d %d %d %d", chan, nickname, numpacks, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR), numqueue, get_int_var(DCC_QUEUE_LIMIT_VAR)))
	{
#endif
		malloc_sprintf(&string, "\037[\037cdcc\037]\037 \002%d\002 file%s offered\037-\037 /ctcp \002%s\002 cdcc send #x for pack #x",
			numpacks, (numpacks == 1) ? "" : "s", nickname);
		send_text(chan, string, NULL);		
		malloc_sprintf(&string, "\037[\037cdcc\037]\037 dcc block size\037:\037 \002%d\002, slots open\037:\037 \002%2d\002/\002%2d\002, dcc queue\037:\037 \002%2d\002/\002%2d\002",
			(blocksize) ? blocksize : 1024, maxdccs - dcc_active_count,
			maxdccs, maxqueue - numqueue, maxqueue);
		send_text(chan, string, NULL);
#ifdef BITCHX
	}
#endif	
	for (ptr = offerlist; ptr; ptr = ptr->next) {
	 	if (ptr->size / 1024 > 999)
			sprintf(size, "\002%3.2f\002mb",
				(float) (ptr->size / 1024) / 1024);
		else
			sprintf(size, "\002%3.2f\002kb", (float) ptr->size / 1024); 
				
#ifdef BITCHX
		if (do_hook(CDCC_PACK_LIST, "%s %d %d %lu %d %s", 
			chan, ptr->num, ptr->numfiles, ptr->size, ptr->gets, ptr->desc))
		{
#endif
			malloc_sprintf(&string, "\037%%\037 #%-2d \037(\037%10s\037:\037\002%4d\002 get%s\037)\037 %-39.39s",
				ptr->num, size, ptr->gets, ptr->gets == 1 ? " " : "s", 
				ptr->desc ? ptr->desc : "no description");
			send_text(chan, string, NULL);
#ifdef BITCHX
		}
#endif
	}
#ifdef BITCHX
	(void) do_hook(CDCC_POSTPACK_LIST, "%s %s %d %d %d %d %d", chan, nickname, numpacks, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR), numqueue, get_int_var(DCC_QUEUE_LIMIT_VAR));
#endif
				
	new_free(&chan);
	new_free(&string);
	return 0;
}
		
/* display CDCC help */
int l_help(char *args, char *rest)
{
	int i;
	
	if (args) {
		for (i = 0; *local[i].name; i++) {
			if (!my_stricmp(args, local[i].name)) {
				put_it("*** \002CDCC\002: %-6s - %s", local[i].name, local[i].help);
				return 0;
			}
		}
		put_it("*** \002CDCC\002: unknown command \002%s\002", args);
		return 0;
	} else 
		for (i = 0; *local[i].name; i++)
			put_it("*** \002CDCC\002: %-6s - %s", local[i].name, local[i].help);
	
	return 0;
}

/* notify the current channel that packs are offered */
int l_notice(char *args, char *rest)
{
	char *string = NULL, *chan = NULL;
	
	if (!get_channel_by_refnum(0) || !numpacks) {
		put_it("*** \002CDCC\002: you %s",
			numpacks ? "are not on a channel!" :
				   "have no packs offered!");
		return 0;
	}
	
	malloc_strcpy(&chan, get_channel_by_refnum(0));
	
#ifdef BITCHX
	if (do_hook(CDCC_PREPACK_LIST, "%s %s %d %d %d %d %d", chan, nickname, numpacks, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR), numqueue, get_int_var(DCC_QUEUE_LIMIT_VAR)))
	{
#endif
		malloc_sprintf(&string, "\037[\037cdcc\037]\037 \002%d\002 file%s offered\037-\037 \037\"\037/ctcp \002%s\002 cdcc list\037\"\037 for pack list",   
			numpacks, (numpacks == 1) ? "" : "s", nickname);
		send_text(chan, string, NULL);
	
#ifdef BITCHX
	}
#endif	
#ifdef BITCHX
	(void) do_hook(CDCC_POSTPACK_LIST, "%s %s %d %d %d %d %d", chan, nickname, numpacks, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR), numqueue, get_int_var(DCC_QUEUE_LIMIT_VAR));
#endif
	new_free(&chan);
	new_free(&string);
	return 0;		
}
		
/* display entries in your send queue */
int l_queue(char *args, char *rest)
{
	queue *ptr;
	int num = 1;
	
	if (!queuelist) {
		put_it("*** \002CDCC\002: there are no queue entries");
		return 0;
	}

	for (ptr = queuelist; ptr; ptr = ptr->next, num++) {
#ifdef BITCHX
		if (do_hook(CDCC_QUEUE_LIST, "%s %s %s %d %d", ptr->nick, my_ctime(ptr->time), ptr->desc, ptr->num, ptr->numfiles))
#endif
			put_it("#\002%2d\002 nick: \037%9s\037 (\002%2d\002 file%s)",
				num, ptr->nick, ptr->numfiles,
				ptr->numfiles == 1 ? " " : "s");
	}
	
	return 0;
} 

/* save all of your offered packs */
int l_save(char *args, char *rest)
{
	FILE *file;
	char *name, *expand;
	pack *ptr;
	int count = 0;
	
	if (!offerlist) {
		put_it("*** \002CDCC\002: you have no packs offered");
		return 0;
	}
	
	expand = expand_twiddle("~");
	name = new_malloc(BIG_BUFFER_SIZE + 1);
	strcpy(name, expand);
	strcat(name, "/cdcc.save");
	
	new_free(&expand);

	if (!(file = fopen(name, "wt"))) {
		put_it("*** \002CDCC\002: couldn't open \"%s\"", name);
		new_free(&name);
		return 0;
	}
	
	for (ptr = offerlist; ptr; ptr = ptr->next) {
		fprintf(file, "%s\n", ptr->file);
		fprintf(file, "%s %d\n", ptr->desc, ptr->numfiles);
		fprintf(file, "%d %ld\n", ptr->gets, ptr->size);
		count++;
	}
		
	fclose(file);
	put_it("*** \002CDCC\002: \002%d\002 pack%s saved to \"cdcc.save\"",
		count, count == 1 ? "" : "s");
	new_free(&name);

	return 0;
} 

/* load packs from cdcc.save */
int l_load(char *args, char *rest)
{
	FILE *file;
	char *buffer, *expand, *temp;
	pack *ptr, *last = NULL;
	int count = 0;
	
	expand = expand_twiddle("~");
	buffer = new_malloc(BIG_BUFFER_SIZE + 1);
	strcpy(buffer, expand);
	strcat(buffer, "/cdcc.save");
	
	new_free(&expand);

	if (!(file = fopen(buffer, "rt"))) {
		put_it("*** \002CDCC\002: couldn't open \"%s\"", buffer);
		new_free(&buffer);
		return 0;
	}

	for (ptr = offerlist; ptr; ptr = ptr->next)
		last = ptr;
	
	while (fgets(buffer, BIG_BUFFER_SIZE, file)) {
		ptr = (pack *) new_malloc(sizeof(pack));
		ptr->num = ++numpacks;
		buffer[strlen(buffer)-1] = 0;
		malloc_strcpy(&ptr->file, buffer);
		fgets(buffer, BIG_BUFFER_SIZE, file);
		buffer[strlen(buffer)-1] = 0;
		temp = rindex(buffer, ' ');
		*temp++ = '\0';
		malloc_strcpy(&ptr->desc, buffer);
		ptr->numfiles = atoi(temp);
		fgets(buffer, BIG_BUFFER_SIZE, file);
		buffer[strlen(buffer)-1] = 0;
		ptr->gets = atoi(strtok(buffer, " "));
		ptr->size = atol(strtok(NULL, "\r"));		
		ptr->next = NULL;
		
		if (last) {
			last->next = ptr;
			last = ptr;
		} else {
			offerlist = ptr;
			last = offerlist;
		}
		count++;
	}
		
	fclose(file);
	put_it("*** \002CDCC\002: \002%d\002 pack%s loaded", count, count == 1 ? "" : "s");
	new_free(&buffer);

	return 0;
} 
	
	
 
/* --- Misc functions --- */

/* add file/files to a pack */
void add_files(char *args, char *rest)
{
	char *thefile, *expand, *path;
	char *temp, *filebuf;
	DIR *dptr;
	struct dirent *dir;
	struct stat statbuf;

	path = new_malloc(BIG_BUFFER_SIZE + 1);
	temp = new_malloc(BIG_BUFFER_SIZE + 1);
	/* I hope this will be enough... 2k of filenames is alot though :) */
	filebuf = new_malloc(2048);

	strcpy(path, rest);
	if ((thefile = strrchr(path, '/')))
		*thefile++ = '\0';
	else {
		strcpy(path, "~");
		thefile = rest;
	}
		
	expand = expand_twiddle(path);
	dptr = opendir(expand);
	if (!dptr) {
		put_it("*** \002CDCC\002: you cannot access dir %s", expand);
		new_free(&expand);
		new_free(&path);
		new_free(&filebuf);
		new_free(&temp); 
		return;
	}
	newpack = (pack *) new_malloc(sizeof(pack));
	while ((dir = readdir(dptr))) {
		if (!dir->d_ino || !wild_match(thefile, dir->d_name))
			continue;
		sprintf(temp, "%s/%s", expand, dir->d_name);
		strcat(filebuf, " ");
		strcat(filebuf, temp);
		stat(temp, &statbuf);
		if (S_ISDIR(statbuf.st_mode))
			continue;
		newpack->size += statbuf.st_size;
		newpack->numfiles++;
	}
	if (!newpack->numfiles) {
		put_it("*** \002CDCC\002: no files found, aborting...");
		new_free(&expand);
		new_free(&path);
		new_free(&filebuf);
		new_free(&temp); 
		new_free((char **) &newpack);
		return;
	}
	
	newpack->num = ++numpacks;
	newpack->file = new_malloc(strlen(filebuf) + 1);
	strcpy(newpack->file, filebuf);

	add_wait_prompt("description of pack: ", add_desc, "", WAIT_PROMPT_LINE);
	new_free(&expand);
	new_free(&filebuf);
	new_free(&path);
	new_free(&temp);
	
	return;
}

/* add a description to the new pack, and add to list */
void add_desc(char *args, char *rest)
{
	pack *ptr, *last = NULL; 
	char size[20];
	
	malloc_strcpy(&newpack->desc, rest);

	for (ptr = offerlist; ptr; ptr = ptr->next)
		last = ptr;

	if (last)
		last->next = newpack;
	else
		offerlist = newpack;
	newpack->next = NULL;
			
	if (newpack->size / 1024 > 999)
		sprintf(size, "\002%3.2f\002mb",
		(double) (newpack->size / 1024) / 1024);
	else
		sprintf(size, "\002%3.2f\002kb",
			(double) newpack->size / 1024); 
	put_it("*** \002CDCC\002: added pack #\002%d\002, \002%d\002 file%s (%s)",
		newpack->num, newpack->numfiles,
		(newpack->numfiles == 1) ? "" : "s", size);

	return;
}

/* handle the actual removing of packs / all packs */
void del_pack(char *args, char *rest)
{
	pack *ptr, *last = offerlist;
	int packnum;
		
	if (!rest || !*rest)
		return;

	if (*rest == '*') {
		numpacks = 0;
		for (ptr = last = offerlist; last;) {
			ptr = last->next;
			new_free(&last->desc);
			new_free(&last->file);
			new_free((char **) &last);
			last = ptr;
		}
		offerlist = NULL;
		put_it("*** \002CDCC\002: removed all packs from offer list");
		return;
	}
	
	packnum = atoi(rest);
	if (!packnum || packnum > numpacks) {
		put_it("*** \002CDCC\002: invalid pack specification", rest);
		return;
	}

	for (ptr = offerlist; ptr; ptr = ptr->next) {
		if (ptr->num == packnum) {
			pack *ptr2;
			if (ptr != offerlist)
				last->next = ptr->next;
			else
				offerlist = ptr->next;

			new_free(&ptr->desc);
			new_free(&ptr->file);
			new_free((char **) &ptr);
			numpacks--;

			for (ptr2 = last->next; ptr2; ptr2 = ptr2->next)
				ptr2->num--;
			put_it("*** \002CDCC\002: removed pack \002%d\002 from offer list", packnum);	
			return;
		}
		last = ptr;
	}
	put_it("*** \002CDCC\002: pack \002%d\002 does not exist", packnum); 
	return;
}

/* add a person to the dcc send queue */
int add_to_queue(char *nick, pack *sendpack)
{
	queue *ptr, *last = NULL, *new;
	
	new = (queue *) new_malloc(sizeof(queue));
	malloc_strcpy(&new->nick, nick);
	malloc_strcpy(&new->file, sendpack->file);
	malloc_strcpy(&new->desc, sendpack->desc);
	new->time = time(NULL);
	new->numfiles = sendpack->numfiles;
	new->next = NULL;
	sendpack->gets++;
	
	for (ptr = queuelist; ptr; ptr = ptr->next)
		last = ptr;

	if (last)
		last->next = new;
	else
		queuelist = new;
	numqueue++;
	send_to_server("NOTICE %s :\002CDCC\002: all slots full... added you to the queue",
		nick);
	
	return 0;
}
	
/* check queue & send files... called by irc.c io() */
void dcc_sendfrom_queue(void) 
{
	queue *ptr = queuelist;
	char *dccinfo = NULL, *file = NULL, *q, *temp = NULL;

	if (!ptr || ptr->numfiles > (get_int_var(DCC_SEND_LIMIT_VAR) - dcc_active_count))
		return;	

	put_it("*** \002CDCC\002: sending \037%s\037 \002%d\002 file%s from queue",
		ptr->nick, ptr->numfiles,
		ptr->numfiles == 1 ? "" : "s");

	malloc_strcpy(&file, ptr->file);
	q = temp = file;
	for (temp = next_arg(file, &file); temp && *temp; temp = next_arg(file, &file)) {
		malloc_sprintf(&dccinfo, "%s %s", ptr->nick, temp);
		dcc_filesend(dccinfo);
	}
	new_free(&dccinfo);

	queuelist = ptr->next;
	new_free(&ptr->nick);
	new_free(&ptr->file);
	new_free((char **) &ptr);
	new_free(&q);
	numqueue--;
	
	return;
}

static time_t plist_last_time = 0;

void get_ptimer(char *args, char *rest)
{
	if (rest && *rest)
		ptimer = strtoul(rest, NULL, 10);
	if (ptimer)
		bitchsay("Set \002Cdcc\002 ptimer value to %d seconds", ptimer);
	else
	{
		plist_last_time = 0;
		bitchsay("Disabled \002Cdcc\002 ptimer");
	}
}

int l_timer(char *args, char *rest)
{
	if (args && *args)
		get_ptimer(NULL, args);
	else
		add_wait_prompt("Set Cdcc p-timer to (0 to disable): ", get_ptimer, "", WAIT_PROMPT_LINE);
	return 0;
}

void cdcc_timer_offer(void)
{
	time_t timenow = time(NULL);
	if (!offerlist || !ptimer)
		return;
	if (timenow - plist_last_time > ptimer)
	{
		plist_last_time = timenow;
		l_plist(NULL, NULL);
	}
}
