

#include "irc.h"

#include <sys/stat.h>

#ifdef ESIX
# include <lan/net_types.h>
#endif

#include "parse.h"
#include "ircterm.h"
#include "server.h"
#include "chelp.h"
#include "edit.h"
#include "crypt.h"
#include "vars.h"
#include "ircaux.h"
#include "lastlog.h"
#include "window.h"
#include "screen.h"
#include "whois.h"
#include "hook.h"
#include "input.h"
#include "ignore.h"
#include "keys.h"
#include "names.h"
#include "alias.h"
#include "history.h"
#include "funny.h"
#include "ctcp.h"
#include "dcc.h"
#include "translat.h"
#include "output.h"
#include "exec.h"
#include "notify.h"
#include "numbers.h"
#include "status.h"
#include "if.h"
#include "help.h"
#include "stack.h"
#include "queue.h"
#include "timer.h"
#include "list.h"
#include "pluto.h"
#include "userlist.h"
#include "whowas.h"

void nwhois (char *command, char *args, char *subargs)
{
ChannelList *tmp;
NickList *user = NULL;
int once = 0;
char *nick;
int mem_use = 0, max_mem = 0;
	if (args && *args)
	{
		tmp = server_list[from_server].chan_list;
		if (!tmp)
		{
			bitchsay("No Channel");
			return;
		}
		nick = next_arg(args, &args);
		while (tmp)
		{
			for (user = tmp->nicks; user; user = user->next)
			{
				if (wild_match(user->nick, nick) || wild_match(nick, user->nick))
				{
					if (!once)
					{
put_it("Channel    @ v  UL: l   a  p  SL: l      +o   -o   +b   -b  kicks  nicks   pub");
						once++;
					}
put_it("%-10s %c %c    %3d  %2d %2d     %2d    %4d %4d %4d %4d  %5d  %5d %5d",
tmp->channel, user->chanop? '@':' ',user->voice?'v':' ',user->userlist?user->userlist->level:0, 
user->userlist?user->userlist->aop:0,user->userlist?user->userlist->prot:0, user->shitlist?user->shitlist->level: 0,
user->stat_ops,user->stat_dops, user->stat_bans, user->stat_unbans, 
user->stat_kicks, user->stat_nicks, user->stat_pub);
					mem_use = sizeof(NickList) + (user->userlist? sizeof(UserList): 0) + (user->shitlist? sizeof(ShitList):0) + (user->botlist?sizeof(UserList):0);
					if (max_mem < mem_use)
						max_mem = mem_use;
					break;
				}
			}				
			tmp = tmp->next;
		}
		if (once)
			bitchsay("End /NWhoIs list for %s [Mem Usage: %d bytes per chan]", nick, max_mem);
		else
			bitchsay("No such nick %s on any channel", nick);
	}
	else
		bitchsay("Nwhois usage: /Nwhois nick"); 
}

void whowas _((char *command, char *args, char *subargs))
{
	if (!command)
		show_whowas();
	else
		show_wholeft();
}

void do_sk _((char *command, char *args, char *subargs))
{
char *serv = NULL;
char *to = NULL;
int count;
NickList *Nick;
ChannelList *chan;

	if (!args || !*args)
	{
		bitchsay("Usage: /SK [server] <channel>");
		return;
	}
	serv = next_arg(args, &args);
	if (*serv == '#')
	{
		bitchsay("You need to specify a server not a channel");
	}
	if (!(to = next_arg(args, &args)))
		to = get_channel_by_refnum(0);
	if (!to || !is_channel(to))
	{
		not_on_a_channel();
		return;
	}
	if (!are_you_opped(to))
	{
		you_not_opped(to);
		return;
	}
	if ((chan = lookup_channel(to, from_server, CHAN_NOUNLINK)))
	{
		bitchsay("Performing Lame Server kicks on %s", chan->channel);
		for (Nick = chan->nicks, count = 0; Nick; Nick = Nick->next)
		{
			if (Nick->server && wild_match(serv, Nick->server))
			{
				if (!isme(Nick->nick))
				{
					send_to_server("KICK %s %s : Lame Server Kick", to, Nick->nick);
					count++;
				}
			}
		}
		if (count)
			bitchsay("Found %d lamerz on %s from %s", count, to, serv);
	} else
		bitchsay("Error in /sk");
}

int check_args(char *remote_host, char *remote_ip)
{
struct hostent *host;

	host = gethostbyname(remote_host);

	if (host == NULL) 
	{
		unsigned char a,b,c,d,n;
		unsigned char addr[5];
		/* hmm.. perhaps it was a dotted quad entered.. */
		n = sscanf(remote_host,"%u.%u.%u.%u",&a,&b,&c,&d);
		if (n != 4) 
		{
			bitchsay("error: host '%s' not found",remote_host);
			return 0;
		}
		addr[0] = a;
		addr[1] = b;
		addr[2] = c;
		addr[3] = d;
		host = gethostbyaddr(addr,4,AF_INET);
		if (host == NULL) {
			bitchsay("error: host '%s' not found",remote_host);
			return 0;
		}
		sprintf(remote_ip,"%u.%u.%u.%u",a,b,c,d);
	} 
	else 
	{
		sprintf(remote_ip,"%u.%u.%u.%u", (unsigned char) host->h_addr_list[0][0],
			(unsigned char) host->h_addr_list[0][1],
			(unsigned char) host->h_addr_list[0][2],
			(unsigned char) host->h_addr_list[0][3]);
	}
	return 1;
}

int scan(char *remote_host, char *remote_ip, int low_port, int high_port)
{
	int soc;
	struct sockaddr_in addr;
	struct servent *serv;
	int port,rc,addr_len,opt;
	int found = 0;

	for (port = low_port;port <= high_port;port++) 
	{

		soc = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);

		if (soc < 0) 
		{
			bitchsay("error: socket() failed");
			return 0;
		}

		rc = setsockopt(soc,SOL_SOCKET,SO_REUSEADDR,&opt,sizeof(opt));

		addr.sin_family = AF_INET;
		addr.sin_addr.s_addr = inet_addr(remote_ip);
		addr.sin_port = htons(port);

		addr_len = sizeof(addr);
		rc = connect(soc, (struct sockaddr*) &addr, addr_len);

		close(soc);


		if (rc < 0)
			continue;

		serv = getservbyport(htons(port),"tcp");
/*
		else if (scan_type == s_udp)
			serv = getservbyport(htons(port),"udp");
		else
			return;
*/
		found++;
		put_it("port %d (%s) is running",port,(serv == NULL)?"UNKNOWN": serv->s_name);
	}
	if (!found)
		bitchsay("Wonder what hostname that was. No ports found");
	return 1;
}

void findports(char *command, char *args, char *subargs)
{
char *remote_host;
char remote_ip[20];
int low_port = 6660;
int high_port = 7000;
	if (args && *args)
	{
		char *tmp = NULL;
		remote_host = next_arg(args, &args);
		if (args && *args)
		{
			tmp = next_arg(args, &args);
			low_port = strtoul(tmp, NULL, 10);	
			if (args && *args)
			{
				tmp = next_arg(args, &args);
				high_port = strtoul(tmp, NULL, 10);	
			}
			else
				high_port = low_port;
		}
		if (check_args(remote_host, remote_ip))
		{
			bitchsay("Scanning %s's tcp ports %d through %d",remote_host, low_port,high_port);
			scan(remote_host, remote_ip, low_port, high_port);
		}
	} else
		bitchsay("Usage: /findports <hostname> [low port [high port]]");
	return;
}

